import os
from typing import Optional, Dict

import boto3
from boto3 import Session


class Credentials:
    """
    Credentials dataclass for storing AWS credentials.
    """

    def __init__(
            self,
            aws_access_key_id: Optional[str] = os.environ.get('AWS_ACCESS_KEY_ID'),
            aws_secret_access_key: Optional[str] = os.environ.get('AWS_SECRET_ACCESS_KEY'),
            aws_session_token: Optional[str] = os.environ.get('AWS_SESSION_TOKEN'),
            profile_name: Optional[str] = os.environ.get('AWS_PROFILE'),
            region_name: Optional[str] = os.environ.get('AWS_DEFAULT_REGION'),
    ) -> None:
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.profile_name = profile_name
        self.region_name = region_name

    @property
    def boto_session(self) -> Session:
        """
        Property for boto session.

        :return: Boto session instance.
        """
        return boto3.session.Session(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            profile_name=self.profile_name,
            region_name=self.region_name
        )

    @property
    def environ(self) -> Optional[Dict[str, str]]:
        """
        Property for serialized credentials to dict.

        :return: Credentials in a form of a dict.
        """
        return {key: value for key, value in {
            'AWS_ACCESS_KEY_ID': self.aws_access_key_id,
            'AWS_SECRET_ACCESS_KEY': self.aws_secret_access_key,
            'AWS_SESSION_TOKEN': self.aws_session_token,
            'AWS_PROFILE': self.profile_name,
            'AWS_DEFAULT_REGION': self.region_name
        }.items() if value} or None



from functools import lru_cache
from typing import Dict, Optional

from aws_cdk.aws_ssm import StringParameter
from aws_cdk.core import Stack, Construct, CfnOutput
from b_cf_outputs.cf_outputs import CfOutputs

from b_aws_testing_framework import b_aws_testing_framework_version
from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.testing_manager import TestingManager


class TestingStack(Stack):
    """
    A testing stack that can be either inherited or created as an instance.
    Useful for managing outputs.
    """

    def __init__(self, scope: Construct) -> None:
        """
        Constructor.

        :param scope: Parent stack or AWS CDK application.
        """
        self.__scope = scope

        super().__init__(
            scope=scope,
            id=self.name(),
            stack_name=self.name()
        )

        self.b_aws_testing_framework_version_ssm_parameter = StringParameter(
            scope=self,
            id=f'{self.name()}.B.AwsTestingFramework.Version',
            parameter_name=f'{self.name()}.B.AwsTestingFramework.Version',
            string_value=str(b_aws_testing_framework_version)
        )

    def add_output(self, key: str, value: str) -> None:
        """
        Add an output here so it could be accessed in tests later.

        :param key: Output name.
        :param value: Output value.

        :return: No return.
        """
        CfnOutput(
            scope=self,
            id=f'{self.global_prefix()}{self.name()}{key}',
            export_name=f'{self.global_prefix()}{key}',
            value=value,
        )

    @staticmethod
    def global_prefix() -> str:
        """
        Returns an already set global prefix.

        :return: Global prefix.
        """
        return TestingManager.get_global_prefix()

    @staticmethod
    def name() -> str:
        """
        Default name of this testing stack.

        :return: Stack name.
        """
        return f'{TestingStack.global_prefix()}TestingStack'

    """
    Methods that should only be used in tests.
    """

    @classmethod
    def get_output(cls, key: str, credentials: Optional[Credentials] = None) -> str:
        """
        Loads this stack's outputs and creates a corresponding one.

        :param key: Output name (key).
        :param credentials: Optional credentials for AWS API commands.

        :return: Output value.
        """
        return cls.load_outputs_cached(credentials)[f'{cls.global_prefix()}{cls.name()}{key}']

    @classmethod
    @lru_cache(maxsize=None)
    def load_outputs_cached(cls, credentials: Optional[Credentials] = None) -> Dict[str, str]:
        """
        Loads and caches (for better performance) this stack's outputs.

        :param credentials: Optional credentials for AWS API commands.

        :return: All outputs in a form of a dictionary.
        """
        return cls.load_outputs(credentials)

    @classmethod
    def load_outputs(cls, credentials: Optional[Credentials] = None) -> Dict[str, str]:
        """
        Loads this stack's outputs.

        :param credentials: Optional credentials for AWS API commands.

        :return: All outputs in a form of a dictionary.
        """
        credentials = credentials or Credentials()
        return CfOutputs(credentials.boto_session).get_outputs(cls.name())[cls.name()]



import random
import string
import tempfile
import os.path
import os
from abc import ABC, abstractmethod
from typing import Optional, Callable, Any

from b_aws_testing_framework.credentials import Credentials


class BaseTestingManager(ABC):
    """
    Abstract class for testing managers.
    """
    __GLOBAL_PREFIX_PATH: str = f'{tempfile.gettempdir()}/B_AWS_TESTING_FRAMEWORK_GLOBAL_PREFIX.txt'

    def __init__(self, credentials: Credentials) -> None:
        """
        Constructor.
        """
        self.__credentials = credentials

    @property
    def credentials(self) -> Credentials:
        """
        Property for credentials attribute.

        :return: Credentials.
        """
        return self.__credentials

    @staticmethod
    def set_global_prefix(prefix: Optional[str] = None, override: bool = True) -> None:
        """
        Sets a global prefix to be used to name resources.

        :param prefix: An optional prefix. If the value is not given, a random string is generated.
        :param override: Specify whether the prefix should be overridden if its already set.

        :return: No return.
        """
        if (override is False) and (BaseTestingManager.is_global_prefix_set() is True):
            return

        prefix = prefix or ''.join(random.choice(
            string.ascii_lowercase + string.ascii_uppercase,
        ) for _ in range(5))

        with open(BaseTestingManager.__GLOBAL_PREFIX_PATH, 'w') as file:
            file.write(prefix)

    @staticmethod
    def get_global_prefix() -> str:
        """
        Gets a global prefix that is used to name resources.

        :return: Global prefix.
        """
        if not BaseTestingManager.is_global_prefix_set():
            raise ValueError('Global prefix is not set!')

        with open(BaseTestingManager.__GLOBAL_PREFIX_PATH) as file:
            prefix = str(file.read())

        if not prefix:
            raise ValueError('Prefix is empty!')

        return prefix

    @staticmethod
    def delete_global_prefix() -> None:
        """
        Deletes a global prefix.

        :return: No return.
        """
        try:
            os.remove(BaseTestingManager.__GLOBAL_PREFIX_PATH)
        except FileNotFoundError:
            pass

    @staticmethod
    def is_global_prefix_set() -> bool:
        """
        Checks whether global prefix is set.

        :return: True if is set, false otherwise.
        """
        return os.path.isfile(BaseTestingManager.__GLOBAL_PREFIX_PATH)

    @abstractmethod
    def prepare_infrastructure(self, custom_deploy_action: Optional[Callable[[Any], Any]] = None) -> None:
        """
        Function to create and prepare infrastructure for testing.

        :param custom_deploy_action: Each child class implements their own deployment functionality. However,
        a client can override such functionality with a custom deploy action. A custom deploy action should be
        responsible just for the infrastructure creation. The custom deploy action should be a callable
        function that takes a configuration object as a parameter. A configuration object is up to each
        child class to define.

        :return: No return.
        """
        pass

    @abstractmethod
    def destroy_infrastructure(self, custom_destroy_action: Optional[Callable[[Any], Any]] = None) -> None:
        """
        Function to destroy the infrastructure.

        :param custom_destroy_action: Each child class implements their own destroy functionality. However,
        a client can override such functionality with a custom destroy action. A custom destroy action should be
        responsible just for the infrastructure destruction. The custom destroy action should be a callable
        function that takes a configuration object as a parameter. A configuration object is up to each
        child class to define.

        :return: No return.
        """
        pass



from typing import Optional, Dict


class CdkToolConfig:
    """
    Configuration class for aws cdk tool.
    """

    def __init__(
            self,
            cdk_app_path: str,
            destroy_before_preparing: bool = True,
            project_root_path: Optional[str] = None,
            deployment_process_environment: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Constructor.

        :param cdk_app_path: Path to an AWS CDK folder where the CDK app is.
        :param destroy_before_preparing: A flag indicating whether the infrastructure manager should attempt
        to clean (destroy) the infrastructure before creating one. This is useful in cases there are some
        resource leftovers from a previous run.
        :param project_root_path: An absolute path to the root of your whole project. This variable is used
        to add to python path. If you face any import errors - it might be a good idea to set this variable.
        :param deployment_process_environment: A dictionary that represent OS-level environment. This environment
        will be passed to a process that deploys/destroys the CDK infrastructure.
        """
        self.__cdk_app_path = cdk_app_path
        self.__destroy_before_preparing = destroy_before_preparing
        self.__project_root_path = project_root_path
        self.__deployment_process_environment = deployment_process_environment or {}

    @property
    def cdk_app_path(self):
        return self.__cdk_app_path

    @property
    def destroy_before_preparing(self):
        return self.__destroy_before_preparing

    @property
    def project_root_path(self):
        return self.__project_root_path

    @property
    def deployment_process_environment(self):
        return self.__deployment_process_environment



import logging
from typing import Optional, Callable, Any

from botocore.exceptions import ClientError

from b_aws_testing_framework.base_testing_manager import BaseTestingManager
from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cf_testing.cf_tool_config import CfToolConfig
from b_aws_testing_framework.tools.cf_testing.stack_waiter import StackWaiter

logger = logging.getLogger(__name__)


class TestingManager(BaseTestingManager):
    """
    Test manager class which prepares infrastructure for tests.
    After tests are finished, destroys the infrastructure.
    """

    def __init__(self, credentials: Credentials, config: CfToolConfig):
        super().__init__(credentials)

        self.__config = config

    def prepare_infrastructure(self, custom_deploy_action: Optional[Callable[[Any], Any]] = None) -> None:
        """
        Prepares infrastructure to run tests.
        Firstly, the infrastructure is destroyed (if any leftovers exist).
        Secondly, the infrastructure is created.

        :return: No return.
        """
        if custom_deploy_action:
            raise NotImplementedError()

        self.set_global_prefix(override=False)
        self.__destroy_infrastructure()
        self.__create_infrastructure()

    def destroy_infrastructure(self, custom_destroy_action: Optional[Callable[[Any], Any]] = None) -> None:
        """
        Destroys the infrastructure.
        
        :return: No return.
        """
        if custom_destroy_action:
            raise NotImplementedError()

        self.__destroy_infrastructure()
        self.delete_global_prefix()

    """
    Infrastructure functions.
    """

    def __create_infrastructure(self) -> None:
        client = self.credentials.boto_session.client('cloudformation')

        with open(self.__config.cf_template_path, 'r') as file:
            client.create_stack(
                StackName=f'{self.get_global_prefix()}TestStack',
                TemplateBody=file.read(),
                Capabilities=['CAPABILITY_IAM'],
            )

        StackWaiter(f'{self.get_global_prefix()}TestStack').wait()

    def __destroy_infrastructure(self) -> None:
        client = self.credentials.boto_session.client('cloudformation')

        try:
            client.delete_stack(
                StackName=f'{self.get_global_prefix()}TestStack',
            )
        except ClientError as ex:
            # Retrieve code and message from an error.
            code = ex.response['Error']['Code']
            message = ex.response['Error']['Message']

            # If stack does not exist, its fine, just return None.
            if code == 'ValidationError' and 'does not exist' in message.lower():
                return

            # Otherwise throw an exception.
            raise

        StackWaiter(f'{self.get_global_prefix()}TestStack').wait()



import logging

from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cf_testing.testing_manager import TestingManager

logger = logging.getLogger(__name__)


def test_stack_exists() -> None:
    """
    Tests that stack actually was created.

    :return: No return.
    """
    # The stack name is provided by the cf testing manager.
    stack_name = f'{TestingManager.get_global_prefix()}TestStack'

    client = Credentials().boto_session.client('cloudformation')

    stacks = client.list_stacks(StackStatusFilter=['CREATE_COMPLETE'])['StackSummaries']
    stacks = [stack['StackName'] for stack in stacks]

    assert stack_name in stacks



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

with open('VERSION') as file:
    VERSION = file.read()
    VERSION = ''.join(VERSION.split())

setup(
    name='b_aws_testing_framework',
    version=VERSION,
    license='Apache License 2.0',
    packages=find_packages(exclude=[
        # Exclude virtual environment.
        'venv',
        # Exclude test source files.
        'b_aws_testing_framework_test'
    ]),
    description=(
        'AWS infrastructure testing framework that supports multiple IAC tools.'
    ),
    long_description=README + '\n\n' + HISTORY,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=[
        'aws-cdk.core>=1.50.0,<2.0.0',
        'aws-cdk.aws-ssm>=1.50.0,<2.0.0',
        'boto3>=1.16.0,<2.0.0',
        'pytest>=6.0.2,<7.0.0',
        'b-continuous-subprocess>=0.3.2,<1.0.0',
        'b-cf-outputs>=0.0.3,<1.0.0'
    ],
    author='Laimonas Sutkus',
    author_email='laimonas.sutkus@biomapas.com',
    keywords='AWS Test IAC',
    url='https://github.com/biomapas/B.AwsTestingFramework.git',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
)



from os.path import abspath as apat
from os.path import dirname as dirn

from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.cdk_tool_config import CdkToolConfig
from b_aws_testing_framework.tools.cdk_testing.testing_manager import TestingManager

CDK_PATH = dirn(apat(__file__))
MANAGER = TestingManager(Credentials(), CdkToolConfig(CDK_PATH))


def pytest_sessionstart(*args, **kwargs):
    """
    Called after command line options have been parsed and
    all plugins and initial conftest files been loaded.
    """
    MANAGER.set_global_prefix()
    MANAGER.prepare_infrastructure()


def pytest_sessionfinish(*args, **kwargs):
    """
    Called before test process is exited.
    """
    MANAGER.destroy_infrastructure()



from aws_cdk.core import Construct

from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack


class Infrastructure(TestingStack):
    """
    This is an entry point for your infrastructure.
    Create other resources and stacks you want to test here.
    """

    def __init__(self, scope: Construct):
        super().__init__(scope=scope)

        self.add_output('TestKey1', 'TestValue1')
        self.add_output('TestKey2', 'TestValue2')
        self.add_output('TestKey3', 'TestValue3')
